# pyenchant
#
# Copyright (C) 2004-2008, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#

"""

enchant._enchant:  ctypes-based wrapper for enchant C library

This module implements the low-level interface to the underlying
C library for enchant.  The interface is based on ctypes and tries
to do as little as possible while making the higher-level components
easier to write.

The following conveniences are provided that differ from the underlying
C API:

    * the "enchant" prefix has been removed from all functions, since
      python has a proper module system
    * callback functions do not take a user_data argument, since
      python has proper closures that can manage this internally
    * string lengths are not passed into functions such as dict_check,
      since python strings know how long they are

"""

import ctypes
import ctypes.util
import os
import os.path
import platform
import sys
import textwrap
from ctypes import CFUNCTYPE, POINTER, c_char_p, c_int, c_size_t, c_void_p, pointer
from typing import Callable, List, Optional, TypeVar  # noqa F401


def from_prefix(prefix: str) -> str:
    find_message("finding from prefix ", prefix)
    assert os.path.exists(prefix), prefix + "  does not exist"
    bin_path = os.path.join(prefix, "bin")
    enchant_dll_path = os.path.join(bin_path, "libenchant-2.dll")
    assert os.path.exists(enchant_dll_path), enchant_dll_path + " does not exist"
    # Make sure all the dlls found next to libenchant-2.dll
    # (libglib-2.0-0.dll, libgmodule-2.0-0.dll, ...) can be
    # used without having to modify %PATH%
    new_path = bin_path + os.pathsep + os.environ["PATH"]
    find_message("Prepending ", bin_path, " to %PATH%")
    os.environ["PATH"] = new_path
    return enchant_dll_path


def from_env_var(library_path: str) -> str:
    find_message("using PYENCHANT_LIBRARY_PATH env var")
    assert os.path.exists(library_path), library_path + " does not exist"
    return library_path


def from_package_resources() -> Optional[str]:
    if sys.platform != "win32":
        return None
    bits, _ = platform.architecture()
    if bits == "64bit":
        subdir = "mingw64"  # hopefully this is compatible
    else:
        subdir = "mingw32"  # ditto
    this_path = os.path.dirname(os.path.abspath(__file__))
    data_path = os.path.join(this_path, "data", subdir)
    find_message("looking in ", data_path)
    if os.path.exists(data_path):
        return from_prefix(data_path)
    return None


def from_system() -> Optional[str]:
    # Note: keep enchant-2 first
    find_message("looking in system")
    candidates = [
        "enchant-2",
        "libenchant-2",
        "enchant",
        "libenchant",
        "enchant-1",
        "libenchant-1",
    ]

    # On M1 macs, Homebrew is in /opt, which isn't on the default
    # DYLD_LIBRARY_PATH. Look for the homebrew path explicitly.
    if sys.platform == "darwin" and platform.machine() == "arm64":
        candidates.extend(
            [f"/opt/homebrew/lib/{candidate}" for candidate in candidates]
        )

    for name in candidates:
        find_message("with name ", name)
        res = ctypes.util.find_library(name)
        if res:
            return res
    return None


VERBOSE_FIND = False


def find_message(*args: str) -> None:
    if not VERBOSE_FIND:
        return
    print("pyenchant:: ", *args, sep="")


def find_c_enchant_lib() -> Optional[str]:
    verbose = os.environ.get("PYENCHANT_VERBOSE_FIND")
    if verbose:
        global VERBOSE_FIND
        VERBOSE_FIND = True
    prefix = os.environ.get("PYENCHANT_ENCHANT_PREFIX")
    if prefix:
        return from_prefix(prefix)

    library_path = os.environ.get("PYENCHANT_LIBRARY_PATH")
    if library_path:
        return from_env_var(library_path)

    from_package = from_package_resources()
    if from_package:
        return from_package

    # Last chance
    return from_system()


enchant_lib_path = find_c_enchant_lib()

if enchant_lib_path is None:
    msg = textwrap.dedent(
        """\
        The 'enchant' C library was not found and maybe needs to be installed.
        See  https://pyenchant.github.io/pyenchant/install.html
        for details
        """
    )
    raise ImportError(msg)


find_message("loading library ", enchant_lib_path)
e = ctypes.cdll.LoadLibrary(enchant_lib_path)

# Always assume the found enchant C dll is inside
# the correct directory layout
prefix_dir = os.path.dirname(os.path.dirname(enchant_lib_path))
if hasattr(e, "enchant_set_prefix_dir") and prefix_dir:
    find_message("setting prefix ", prefix_dir)
    e.enchant_set_prefix_dir(prefix_dir.encode())


def callback(restype, *argtypes):
    """Factory for generating callback function prototypes.

    This is factored into a factory so I can easily change the definition
    for experimentation or debugging.
    """
    return CFUNCTYPE(restype, *argtypes)


t_broker_desc_func = callback(None, c_char_p, c_char_p, c_char_p, c_void_p)
t_dict_desc_func = callback(None, c_char_p, c_char_p, c_char_p, c_char_p, c_void_p)


# Simple typedefs for readability

t_broker = c_void_p
t_dict = c_void_p
_B = TypeVar("_B", bound=t_broker)
_D = TypeVar("_D", bound=t_dict)


# Now we can define the types of each function we are going to use

broker_init = e.enchant_broker_init
broker_init.argtypes = []
broker_init.restype = t_broker

broker_free = e.enchant_broker_free
broker_free.argtypes = [t_broker]
broker_free.restype = None

broker_request_dict = e.enchant_broker_request_dict
broker_request_dict.argtypes = [t_broker, c_char_p]
broker_request_dict.restype = t_dict

broker_request_pwl_dict = e.enchant_broker_request_pwl_dict
broker_request_pwl_dict.argtypes = [t_broker, c_char_p]
broker_request_pwl_dict.restype = t_dict

broker_free_dict = e.enchant_broker_free_dict
broker_free_dict.argtypes = [t_broker, t_dict]
broker_free_dict.restype = None

broker_dict_exists = e.enchant_broker_dict_exists
broker_dict_exists.argtypes = [t_broker, c_char_p]
broker_dict_exists.restype = c_int

broker_set_ordering = e.enchant_broker_set_ordering
broker_set_ordering.argtypes = [t_broker, c_char_p, c_char_p]
broker_set_ordering.restype = None

broker_get_error = e.enchant_broker_get_error
broker_get_error.argtypes = [t_broker]
broker_get_error.restype = c_char_p

broker_describe1 = e.enchant_broker_describe
broker_describe1.argtypes = [t_broker, t_broker_desc_func, c_void_p]
broker_describe1.restype = None


def broker_describe(broker: _B, cbfunc) -> None:
    def cbfunc1(*args):
        cbfunc(*args[:-1])

    broker_describe1(broker, t_broker_desc_func(cbfunc1), None)


broker_list_dicts1 = e.enchant_broker_list_dicts
broker_list_dicts1.argtypes = [t_broker, t_dict_desc_func, c_void_p]
broker_list_dicts1.restype = None


def broker_list_dicts(broker: _B, cbfunc) -> None:
    def cbfunc1(*args):
        cbfunc(*args[:-1])

    broker_list_dicts1(broker, t_dict_desc_func(cbfunc1), None)


try:
    broker_get_param = e.enchant_broker_get_param  # type: Callable[[_B, bytes], bytes]
except AttributeError:
    #  Make the lookup error occur at runtime
    def broker_get_param(broker: _B, name: bytes) -> bytes:
        return e.enchant_broker_get_param(broker, name)

else:
    broker_get_param.argtypes = [t_broker, c_char_p]  # type: ignore
    broker_get_param.restype = c_char_p  # type: ignore

try:
    broker_set_param = (
        e.enchant_broker_set_param
    )  # type: Callable[[_B, bytes, bytes], None]
except AttributeError:
    #  Make the lookup error occur at runtime
    def broker_set_param(broker: _B, name: bytes, value: bytes) -> None:
        return e.enchant_broker_set_param(broker, name, value)

else:
    broker_set_param.argtypes = [t_broker, c_char_p, c_char_p]  # type: ignore
    broker_set_param.restype = None  # type: ignore

try:
    get_version = e.enchant_get_version  # type: Callable[[], bytes]
except AttributeError:
    #  Make the lookup error occur at runtime
    def get_version() -> bytes:
        return e.enchant_get_version()

else:
    get_version.argtypes = []  # type: ignore
    get_version.restype = c_char_p  # type: ignore

try:
    set_prefix_dir = e.enchant_set_prefix_dir  # type: Callable[[bytes], None]
except AttributeError:
    #  Make the lookup error occur at runtime
    def set_prefix_dir(path: bytes):
        return e.enchant_set_prefix_dir(path)

else:
    set_prefix_dir.argtypes = [c_char_p]  # type: ignore
    set_prefix_dir.restype = None  # type: ignore

try:
    get_user_config_dir = e.enchant_get_user_config_dir  # type: Callable[[], bytes]
except AttributeError:
    #  Make the lookup error occur at runtime
    def get_user_config_dir() -> bytes:
        return e.enchant_get_user_config_dir()

else:
    get_user_config_dir.argtypes = []  # type: ignore
    get_user_config_dir.restype = c_char_p  # type: ignore

dict_check1 = e.enchant_dict_check
dict_check1.argtypes = [t_dict, c_char_p, c_size_t]
dict_check1.restype = c_int


def dict_check(dict: _D, word: bytes) -> int:
    return dict_check1(dict, word, len(word))


dict_suggest1 = e.enchant_dict_suggest
dict_suggest1.argtypes = [t_dict, c_char_p, c_size_t, POINTER(c_size_t)]
dict_suggest1.restype = POINTER(c_char_p)


def dict_suggest(dict: _D, word: bytes) -> List[bytes]:
    num_suggs_p = pointer(c_size_t(0))
    suggs_c = dict_suggest1(dict, word, len(word), num_suggs_p)
    suggs = []
    n = 0
    while n < num_suggs_p.contents.value:
        suggs.append(suggs_c[n])
        n = n + 1
    if num_suggs_p.contents.value > 0:
        dict_free_string_list(dict, suggs_c)
    return suggs


dict_add1 = e.enchant_dict_add
dict_add1.argtypes = [t_dict, c_char_p, c_size_t]
dict_add1.restype = None


def dict_add(dict: _D, word: bytes) -> None:
    return dict_add1(dict, word, len(word))


dict_add_to_pwl1 = e.enchant_dict_add
dict_add_to_pwl1.argtypes = [t_dict, c_char_p, c_size_t]
dict_add_to_pwl1.restype = None


def dict_add_to_pwl(dict: _D, word: bytes) -> None:
    return dict_add_to_pwl1(dict, word, len(word))


dict_add_to_session1 = e.enchant_dict_add_to_session
dict_add_to_session1.argtypes = [t_dict, c_char_p, c_size_t]
dict_add_to_session1.restype = None


def dict_add_to_session(dict: _D, word: bytes) -> None:
    return dict_add_to_session1(dict, word, len(word))


dict_remove1 = e.enchant_dict_remove
dict_remove1.argtypes = [t_dict, c_char_p, c_size_t]
dict_remove1.restype = None


def dict_remove(dict: _D, word: bytes) -> None:
    return dict_remove1(dict, word, len(word))


dict_remove_from_session1 = e.enchant_dict_remove_from_session
dict_remove_from_session1.argtypes = [t_dict, c_char_p, c_size_t]
dict_remove_from_session1.restype = c_int


def dict_remove_from_session(dict: _D, word: bytes) -> int:
    return dict_remove_from_session1(dict, word, len(word))


dict_is_added1 = e.enchant_dict_is_added
dict_is_added1.argtypes = [t_dict, c_char_p, c_size_t]
dict_is_added1.restype = c_int


def dict_is_added(dict: _D, word: bytes) -> int:
    return dict_is_added1(dict, word, len(word))


dict_is_removed1 = e.enchant_dict_is_removed
dict_is_removed1.argtypes = [t_dict, c_char_p, c_size_t]
dict_is_removed1.restype = c_int


def dict_is_removed(dict: _D, word: bytes) -> int:
    return dict_is_removed1(dict, word, len(word))


dict_store_replacement1 = e.enchant_dict_store_replacement
dict_store_replacement1.argtypes = [t_dict, c_char_p, c_size_t, c_char_p, c_size_t]
dict_store_replacement1.restype = None


def dict_store_replacement(dict: _D, mis: bytes, cor: bytes) -> None:
    return dict_store_replacement1(dict, mis, len(mis), cor, len(cor))


dict_free_string_list = e.enchant_dict_free_string_list
dict_free_string_list.argtypes = [t_dict, POINTER(c_char_p)]
dict_free_string_list.restype = None

dict_get_error = e.enchant_dict_get_error
dict_get_error.argtypes = [t_dict]
dict_get_error.restype = c_char_p

dict_describe1 = e.enchant_dict_describe
dict_describe1.argtypes = [t_dict, t_dict_desc_func, c_void_p]
dict_describe1.restype = None


def dict_describe(dict: _D, cbfunc) -> None:
    def cbfunc1(tag, name, desc, file, data):
        cbfunc(tag, name, desc, file)

    dict_describe1(dict, t_dict_desc_func(cbfunc1), None)



# pyenchant
#
# Copyright (C) 2004-2008, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

enchant.tokenize.en:    Tokenizer for the English language

This module implements a PyEnchant text tokenizer for the English
language, based on very simple rules.

"""

import unicodedata
from typing import Any, Callable, Container, Union  # noqa F401

import enchant.tokenize

_TextLike = Union[bytes, str, bytearray]
_BinaryLike = Union[bytes, bytearray]


class tokenize(enchant.tokenize.tokenize):  # noqa: N801
    """Iterator splitting text into words, reporting position.

    This iterator takes a text string as input, and yields tuples
    representing each distinct word found in the text.  The tuples
    take the form:

        (<word>,<pos>)

    Where `word` is the word string found and `pos` is the position
    of the start of the word within the text.

    The optional argument `valid_chars` may be used to specify a
    list of additional characters that can form part of a word.
    By default, this list contains only the apostrophe ('). Note that
    these characters cannot appear at the start or end of a word.
    """

    _DOC_ERRORS = ["pos", "pos"]

    def __init__(self, text: _TextLike, valid_chars: Container[str] = None) -> None:
        self._valid_chars = valid_chars  # type: Container[str] # type: ignore
        self._text = text  # type: ignore
        self._offset = 0
        # Select proper implementation of self._consume_alpha.
        # 'text' isn't necessarily a string (it could be e.g. a mutable array)
        # so we can't use isinstance(text, str) to detect unicode.
        # Instead we typetest the first character of the text.
        # If there's no characters then it doesn't matter what implementation
        # we use since it won't be called anyway.
        try:
            char1 = text[0]
        except IndexError:
            self._initialize_for_binary()
        else:
            if isinstance(char1, str):
                self._initialize_for_unicode()
            else:
                self._initialize_for_binary()

    def _initialize_for_binary(self) -> None:
        self._consume_alpha = self._consume_alpha_b  # type: Callable[[Any, int], int]
        if self._valid_chars is None:
            self._valid_chars = ("'",)

    def _initialize_for_unicode(self) -> None:
        self._consume_alpha = self._consume_alpha_u
        if self._valid_chars is None:
            # XXX TODO: this doesn't seem to work correctly with the
            # MySpell provider, disabling for now.
            # Allow unicode typographic apostrophe
            # self._valid_chars = (u"'",u"\u2019")
            self._valid_chars = ("'",)

    def _consume_alpha_b(self, text: _BinaryLike, offset: int) -> int:
        """Consume an alphabetic character from the given bytestring.

        Given a bytestring and the current offset, this method returns
        the number of characters occupied by the next alphabetic character
        in the string.  Non-ASCII bytes are interpreted as utf-8 and can
        result in multiple characters being consumed.
        """
        assert offset < len(text)
        if text[offset] >= 0x80:
            return self._consume_alpha_utf8(text, offset)
        if chr(text[offset]).isalpha():
            return 1
        return 0

    def _consume_alpha_utf8(self, text: _BinaryLike, offset: int) -> int:
        """Consume a sequence of utf8 bytes forming an alphabetic character."""
        incr = 2
        u = ""
        while not u and incr <= 4:
            try:
                try:
                    #  In the common case this will be a string
                    u = text[offset : offset + incr].decode("utf8")
                except AttributeError:
                    #  Looks like it was e.g. a mutable char array.
                    try:
                        s = text[offset : offset + incr].tostring()  # type: ignore
                    except AttributeError:
                        s = bytes(c for c in text[offset : offset + incr])
                    u = s.decode("utf8")
            except UnicodeDecodeError:
                incr += 1
        if not u:
            return 0
        if u.isalpha():
            return incr
        if unicodedata.category(u)[0] == "M":
            return incr
        return 0

    def _consume_alpha_u(self, text: str, offset: int) -> int:
        """Consume an alphabetic character from the given unicode string.

        Given a unicode string and the current offset, this method returns
        the number of characters occupied by the next alphabetic character
        in the string.  Trailing combining characters are consumed as a
        single letter.
        """
        assert offset < len(text)
        incr = 0
        if text[offset].isalpha():
            incr = 1
            while offset + incr < len(text):
                if unicodedata.category(text[offset + incr])[0] != "M":
                    break
                incr += 1
        return incr

    def next(self) -> enchant.tokenize.Token:
        text = self._text
        offset = self._offset
        while offset < len(text):
            # Find start of next word (must be alpha)
            while offset < len(text):
                incr = self._consume_alpha(text, offset)
                if incr:
                    break
                offset += 1
            cur_pos = offset
            # Find end of word using, allowing valid_chars
            while offset < len(text):
                incr = self._consume_alpha(text, offset)
                if not incr:
                    if text[offset] in self._valid_chars:
                        incr = 1
                    else:
                        break
                offset += incr
            # Return if word isn't empty
            if cur_pos != offset:
                # Make sure word doesn't end with a valid_char
                while text[offset - 1] in self._valid_chars:
                    offset = offset - 1
                self._offset = offset
                return (text[cur_pos:offset], cur_pos)
        self._offset = offset
        raise StopIteration()



# pyenchant
#
# Copyright (C) 2022, Nico Gulden, Univention GmbH
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

enchant.tokenize.de:    Tokenizer for the German language

This module implements a PyEnchant text tokenizer for the German
language, based on very simple rules.

"""


from typing import Container, Optional

from enchant.tokenize.en import tokenize as tokenizer_en

from .en import _TextLike


class tokenize(tokenizer_en):  # noqa: N801
    def __init__(
        self, text: _TextLike, valid_chars: Optional[Container[str]] = ("-", ".")
    ) -> None:
        super().__init__(text, valid_chars)



# pyenchant
#
# Copyright (C) 2004-2011, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""
enchant:  Access to the enchant spellchecking library
=====================================================

This module provides several classes for performing spell checking
via the Enchant spellchecking library.  For more details on Enchant,
visit the project website:

    https://abiword.github.io/enchant/

Spellchecking is performed using :py:class:`Dict` objects, which represent
a language dictionary.  Their use is best demonstrated by a quick
example::

    >>> import enchant
    >>> d = enchant.Dict("en_US")   # create dictionary for US English
    >>> d.check("enchant")
    True
    >>> d.check("enchnt")
    False
    >>> d.suggest("enchnt")
    ['enchant', 'enchants', 'enchanter', 'penchant', 'incant', 'enchain', 'enchanted']

Languages are identified by standard string tags such as "en" (English)
and "fr" (French).  Specific language dialects can be specified by
including an additional code - for example, "en_AU" refers to Australian
English.  The later form is preferred as it is more widely supported.

To check whether a dictionary exists for a given language, the function
:py:func:`dict_exists` is available.  Dictionaries may also be created using the
function :py:func:`request_dict`.

A finer degree of control over the dictionaries and how they are created
can be obtained using one or more :py:class:`Broker` objects.  These objects are
responsible for locating dictionaries for a specific language.

Note that unicode strings are expected throughout the entire API.
Bytestrings should not be passed into any function.

Errors that occur in this module are reported by raising subclasses
of :py:exc:`~.errors.Error`.

"""
_DOC_ERRORS = ["enchnt", "enchnt", "incant", "fr"]

__version__ = "3.3.0"

import os
import warnings

try:
    from enchant import _enchant as _e
except ImportError:
    if not os.environ.get("PYENCHANT_IGNORE_MISSING_LIB", False):
        raise
    _e = None  # type: ignore

from typing import Any, List, NoReturn, Optional, Tuple, Type, Union  # noqa F401

from enchant.errors import *  # noqa F401,F403
from enchant.errors import DictNotFoundError, Error
from enchant.pypwl import PyPWL
from enchant.utils import get_default_language


class ProviderDesc:
    """Simple class describing an Enchant provider.

    Each provider has the following information associated with it:

        * :py:attr:`name`:        Internal provider name (e.g. "aspell")
        * :py:attr:`desc`:        Human-readable description (e.g. "Aspell Provider")
        * :py:attr:`file`:        Location of the library containing the provider

    """

    _DOC_ERRORS = ["desc"]

    def __init__(self, name: str, desc: str, file: str) -> None:
        self.name = name
        self.desc = desc
        self.file = file

    def __str__(self) -> str:
        return "<Enchant: %s>" % self.desc

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, pd):
        """Equality operator on ProviderDesc objects."""
        if not isinstance(pd, ProviderDesc):
            return False
        return self.name == pd.name and self.desc == pd.desc and self.file == pd.file

    def __hash__(self):
        """Hash operator on ProviderDesc objects."""
        return hash(self.name + self.desc + self.file)


class _EnchantObject:
    """Base class for enchant objects.

    This class implements some general functionality for interfacing with
    the '_enchant' C-library in a consistent way.  All public objects
    from the 'enchant' module are subclasses of this class.

    All enchant objects have an attribute :py:attr:`_this` which contains the
    pointer to the underlying C-library object.  The method :py:meth:`_check_this`
    can be called to ensure that this point is not None, raising an
    exception if it is.
    """

    def __init__(self) -> None:
        """_EnchantObject constructor."""
        self._this = None
        #  To be importable when enchant C lib is missing, we need
        #  to create a dummy default broker.
        if _e is not None:
            self._init_this()

    def _check_this(self, msg: str = None) -> None:
        """Check that :py:attr:`_this` is set to a pointer, rather than `None`."""
        if self._this is None:
            if msg is None:
                msg = "%s unusable: the underlying C-library object has been freed."
                msg = msg % (self.__class__.__name__,)
            raise Error(msg)

    def _init_this(self) -> None:
        """Initialise the underlying C-library object pointer."""
        raise NotImplementedError

    def _raise_error(
        self, default: str = "Unspecified Error", eclass: Type[Error] = Error
    ) -> NoReturn:
        """Raise an exception based on available error messages.

        This method causes an :py:exc:`~.errors.Error` to be raised.  Subclasses should
        override it to retrieve an error indication from the underlying
        API if possible.  If such a message cannot be retrieved, the
        argument value `default` is used.  The class of the exception
        can be specified using the argument `eclass`
        """
        raise eclass(default)

    _raise_error._DOC_ERRORS = ["eclass"]  # type: ignore

    def __getstate__(self):
        """Customize pickling of PyEnchant objects.

        Since it's not safe for multiple objects to share the same C-library
        object, we make sure it's unset when pickling.
        """
        state = self.__dict__.copy()
        state["_this"] = None
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._init_this()


class Broker(_EnchantObject):
    """Broker object for the Enchant spellchecker.

    `Broker` objects are responsible for locating and managing dictionaries.
    Unless custom functionality is required, there is no need to use `Broker`
    objects directly. The py:mod:`enchant` module provides a default broker object
    so that :py:class:`Dict` objects can be created directly.

    The most important methods of this class include:

        * :py:meth:`dict_exists`:   check existence of a specific language dictionary
        * :py:meth:`request_dict`:  obtain a dictionary for specific language
        * :py:meth:`set_ordering`:  specify which dictionaries to try for a given language.

    """

    def __init__(self) -> None:
        """Broker object constructor.

        This method is the constructor for the `Broker` object.  No
        arguments are required.
        """
        super().__init__()

    def _init_this(self) -> None:
        self._this = _e.broker_init()
        if not self._this:
            raise Error("Could not initialise an enchant broker.")
        self._live_dicts = {}

    def __del__(self) -> None:
        """Broker object destructor."""
        # Calling free() might fail if python is shutting down
        try:
            self._free()
        except (AttributeError, TypeError):
            pass

    def __getstate__(self):
        state = super().__getstate__()
        state.pop("_live_dicts")
        return state

    def _raise_error(
        self, default: str = "Unspecified Error", eclass: Type[Error] = Error
    ) -> NoReturn:
        """Overrides _EnchantObject._raise_error to check broker errors."""
        err = _e.broker_get_error(self._this)
        if err == "" or err is None:
            raise eclass(default)
        raise eclass(err.decode())

    def _free(self) -> None:
        """Free system resource associated with a Broker object.

        This method can be called to free the underlying system resources
        associated with a `Broker` object.  It is called automatically when
        the object is garbage collected.  If called explicitly, the
        `Broker` and any associated `Dict` objects must no longer be used.
        """
        if self._this is not None:
            # During shutdown, this finalizer may be called before
            # some Dict finalizers.  Ensure all pointers are freed.
            for dict, count in list(self._live_dicts.items()):
                while count:
                    self._free_dict_data(dict)
                    count -= 1
            _e.broker_free(self._this)
            self._this = None

    def request_dict(self, tag: str = None) -> "Dict":
        """Request a :py:class:`Dict` object for the language specified by `tag`.

        This method constructs and returns a :py:class:`Dict` object for the
        requested language.  `tag` should be a string of the appropriate
        form for specifying a language, such as "fr" (French) or "en_AU"
        (Australian English).  The existence of a specific language can
        be tested using the method :py:meth:`dict_exists`.

        If `tag` is not given or is `None`, an attempt is made to determine
        the current language in use.  If this cannot be determined, :py:exc:`~.errors.Error`
        is raised.

        .. note::
            this method is functionally equivalent to calling the :py:class:`Dict()`
            constructor and passing in the `broker` argument.

        """
        return Dict(tag, self)

    request_dict._DOC_ERRORS = ["fr"]  # type: ignore

    def _request_dict_data(self, tag: str) -> _e.t_dict:
        """Request raw C pointer data for a dictionary.

        This method call passes on the call to the C library, and does
        some internal bookkeeping.
        """
        self._check_this()
        new_dict = _e.broker_request_dict(self._this, tag.encode())
        if new_dict is None:
            e_str = "Dictionary for language '%s' could not be found\n"
            e_str += "Please check https://pyenchant.github.io/pyenchant/ for details"
            self._raise_error(e_str % (tag,), DictNotFoundError)
        if new_dict not in self._live_dicts:
            self._live_dicts[new_dict] = 1
        else:
            self._live_dicts[new_dict] += 1
        return new_dict

    def request_pwl_dict(self, pwl: str) -> "Dict":
        """Request a Dict object for a personal word list.

        This method behaves as :py:meth:`request_dict` but rather than returning
        a dictionary for a specific language, it returns a dictionary
        referencing a personal word list.  A personal word list is a file
        of custom dictionary entries, one word per line.
        """
        self._check_this()
        new_dict = _e.broker_request_pwl_dict(self._this, pwl.encode())
        if new_dict is None:
            e_str = "Personal Word List file '%s' could not be loaded"
            self._raise_error(e_str % (pwl,))
        if new_dict not in self._live_dicts:
            self._live_dicts[new_dict] = 1
        else:
            self._live_dicts[new_dict] += 1
        d = Dict(False)
        d._switch_this(new_dict, self)
        return d

    def _free_dict(self, dict: "Dict") -> None:
        """Free memory associated with a dictionary.

        This method frees system resources associated with a `Dict` object.
        It is equivalent to calling the object`s `free' method.  Once this
        method has been called on a dictionary, it must not be used again.
        """
        self._free_dict_data(dict._this)
        dict._this = None
        dict._broker = None

    def _free_dict_data(self, dict: _e.t_dict) -> None:
        """Free the underlying pointer for a dict."""
        self._check_this()
        _e.broker_free_dict(self._this, dict)
        self._live_dicts[dict] -= 1
        if self._live_dicts[dict] == 0:
            del self._live_dicts[dict]

    def dict_exists(self, tag: str) -> bool:
        """Check availability of a dictionary.

        This method checks whether there is a dictionary available for
        the language specified by `tag`.  It returns `True` if a dictionary
        is available, and `False` otherwise.
        """
        self._check_this()
        val = _e.broker_dict_exists(self._this, tag.encode())
        return bool(val)

    def set_ordering(self, tag: str, ordering: str) -> None:
        """Set dictionary preferences for a language.

        The Enchant library supports the use of multiple dictionary programs
        and multiple languages.  This method specifies which dictionaries
        the broker should prefer when dealing with a given language.  `tag`
        must be an appropriate language specification and `ordering` is a
        string listing the dictionaries in order of preference.  For example
        a valid ordering might be "aspell,myspell,ispell".
        The value of `tag` can also be set to "*" to set a default ordering
        for all languages for which one has not been set explicitly.
        """
        self._check_this()
        _e.broker_set_ordering(self._this, tag.encode(), ordering.encode())

    def describe(self) -> List[ProviderDesc]:
        """Return list of provider descriptions.

        This method returns a list of descriptions of each of the
        dictionary providers available.  Each entry in the list is a
        :py:class:`ProviderDesc` object.
        """
        self._check_this()
        self.__describe_result = []
        _e.broker_describe(self._this, self.__describe_callback)
        return [ProviderDesc(*r) for r in self.__describe_result]

    def __describe_callback(self, name: bytes, desc: bytes, file: bytes) -> None:
        """Collector callback for dictionary description.

        This method is used as a callback into the _enchant function
        `enchant_broker_describe`.  It collects the given arguments in
        a tuple and appends them to the list :py:attr:`__describe_result`.
        """
        name = name.decode()
        desc = desc.decode()
        file = file.decode()
        self.__describe_result.append((name, desc, file))

    def list_dicts(self) -> List[Tuple[str, ProviderDesc]]:
        """Return list of available dictionaries.

        This method returns a list of dictionaries available to the
        broker.  Each entry in the list is a two-tuple of the form:

            (tag,provider)

        where `tag` is the language lag for the dictionary and
        `provider` is a :py:class:`ProviderDesc` object describing the provider
        through which that dictionary can be obtained.
        """
        self._check_this()
        self.__list_dicts_result = []
        _e.broker_list_dicts(self._this, self.__list_dicts_callback)
        return [(r[0], ProviderDesc(*r[1])) for r in self.__list_dicts_result]

    def __list_dicts_callback(self, tag, name, desc, file):
        """Collector callback for listing dictionaries.

        This method is used as a callback into the _enchant function
        `enchant_broker_list_dicts`.  It collects the given arguments into
        an appropriate tuple and appends them to :py:attr:`__list_dicts_result`.
        """
        tag = tag.decode()
        name = name.decode()
        desc = desc.decode()
        file = file.decode()
        self.__list_dicts_result.append((tag, (name, desc, file)))

    def list_languages(self) -> List[str]:
        """List languages for which dictionaries are available.

        This function returns a list of language tags for which a
        dictionary is available.
        """
        langs = []
        for tag, prov in self.list_dicts():
            if tag not in langs:
                langs.append(tag)
        return langs

    def __describe_dict(self, dict_data):
        """Get the description tuple for a dict data object.
        `dict_data` must be a C-library pointer to an enchant dictionary.
        The return value is a tuple of the form:
                (<tag>,<name>,<desc>,<file>)
        """
        # Define local callback function
        cb_result = []

        def cb_func(tag, name, desc, file):
            tag = tag.decode()
            name = name.decode()
            desc = desc.decode()
            file = file.decode()
            cb_result.append((tag, name, desc, file))

        # Actually call the describer function
        _e.dict_describe(dict_data, cb_func)
        return cb_result[0]

    __describe_dict._DOC_ERRORS = ["desc"]  # type: ignore

    def get_param(self, name: str) -> Any:
        """Get the value of a named parameter on this broker.

        Parameters are used to provide runtime information to individual
        provider backends.  See the method :py:meth:`set_param` for more details.

        .. warning::

            This method does **not** work when using the Enchant C
            library version 2.0 and above
        """
        param = _e.broker_get_param(self._this, name.encode())
        if param is not None:
            param = param.decode()
        return param

    get_param._DOC_ERRORS = ["param"]  # type: ignore

    def set_param(self, name: str, value: Any) -> None:
        """Set the value of a named parameter on this broker.

        Parameters are used to provide runtime information to individual
        provider backends.

        .. warning::

            This method does **not** work when using the Enchant C
            library version 2.0 and above
        """
        name = name.encode()
        if value is not None:
            value = value.encode()
        _e.broker_set_param(self._this, name, value)


class Dict(_EnchantObject):
    """Dictionary object for the Enchant spellchecker.

    Dictionary objects are responsible for checking the spelling of words
    and suggesting possible corrections.  Each dictionary is owned by a
    :py:class:`Broker` object, but unless a new :py:class:`Broker` has explicitly been created
    then this will be the :py:mod:`enchant` module default :py:class:`Broker` and is of little
    interest.

    The important methods of this class include:

        * :py:meth:`check()`:              check whether a word is spelled correctly
        * :py:meth:`suggest()`:            suggest correct spellings for a word
        * :py:meth:`add()`:                add a word to the user's personal dictionary
        * :py:meth:`remove()`:             add a word to the user's personal exclude list
        * :py:meth:`add_to_session()`:     add a word to the current spellcheck session
        * :py:meth:`store_replacement()`:  indicate a replacement for a given word

    Information about the dictionary is available using the following
    attributes:

        * :py:attr:`tag`:        the language tag of the dictionary
        * :py:attr:`provider`:   a :py:class:`ProviderDesc` object for the dictionary provider

    """

    def __init__(
        self, tag: Optional[str] = None, broker: Optional[Broker] = None
    ) -> None:
        """Dict object constructor.

        A dictionary belongs to a specific language, identified by the
        string `tag`.  If the tag is not given or is `None`, an attempt to
        determine the language currently in use is made using the :py:mod:`locale`
        module.  If the current language cannot be determined, :py:exc:`~.errors.Error` is raised.

        If `tag` is instead given the value of `False`, a 'dead' Dict object
        is created without any reference to a language.  This is typically
        only useful within PyEnchant itself.  Any other non-string value
        for `tag` raises :py:exc:`~.errors.Error`.

        Each dictionary must also have an associated `Broker` object which
        obtains the dictionary information from the underlying system. This
        may be specified using `broker`.  If not given, the default broker
        is used.
        """
        # Initialise misc object attributes to None
        self.provider = None
        # If no tag was given, use the default language
        if tag is None:
            tag = get_default_language()
            if tag is None:
                raise Error(
                    "No tag specified and default language could not be determined."
                )
        self.tag = tag
        # If no broker was given, use the default broker
        if broker is None:
            broker = _broker
        self._broker = broker
        # Now let the superclass initialise the C-library object
        super().__init__()

    def _init_this(self) -> None:
        # Create dead object if False was given as the tag.
        # Otherwise, use the broker to get C-library pointer data.
        self._this = None
        if self.tag:
            this = self._broker._request_dict_data(self.tag)
            self._switch_this(this, self._broker)

    def __del__(self) -> None:
        """Dict object destructor."""
        # Calling free() might fail if python is shutting down
        try:
            self._free()
        except (AttributeError, TypeError):
            pass

    def _switch_this(self, this, broker: Broker) -> None:
        """Switch the underlying C-library pointer for this object.

        As all useful state for a `Dict` is stored by the underlying C-library
        pointer, it is very convenient to allow this to be switched at
        run-time.  Pass a new dict data object into this method to affect
        the necessary changes.  The creating `Broker` object (at the Python
        level) must also be provided.

        This should *never* *ever* be used by application code.  It's
        a convenience for developers only, replacing the clunkier `data`
        parameter to `__init__` from earlier versions.
        """
        # Free old dict data
        Dict._free(self)
        # Hook in the new stuff
        self._this = this
        self._broker = broker
        # Update object properties
        desc = self.__describe(check_this=False)
        self.tag = desc[0]
        self.provider = ProviderDesc(*desc[1:])

    _switch_this._DOC_ERRORS = ["init"]  # type: ignore

    def _check_this(self, msg: Optional[str] = None) -> None:
        """Extend `_EnchantObject._check_this()` to check Broker validity.

        It is possible for the managing Broker object to be freed without
        freeing the `Dict`.  Thus validity checking must take into account
        `self._broker._this` as well as `self._this`.
        """
        if self._broker is None or self._broker._this is None:
            self._this = None
        super()._check_this(msg)

    def _raise_error(
        self, default: str = "Unspecified Error", eclass: Type[Error] = Error
    ) -> NoReturn:
        """Overrides `_EnchantObject._raise_error` to check dict errors."""
        err = _e.dict_get_error(self._this)
        if err == "" or err is None:
            raise eclass(default)
        raise eclass(err.decode())

    def _free(self) -> None:
        """Free the system resources associated with a `Dict` object.

        This method frees underlying system resources for a `Dict` object.
        Once it has been called, the `Dict` object must no longer be used.
        It is called automatically when the object is garbage collected.
        """
        if self._this is not None:
            # The broker may have been freed before the dict.
            # It will have freed the underlying pointers already.
            if self._broker is not None and self._broker._this is not None:
                self._broker._free_dict(self)

    def check(self, word: str) -> bool:
        """Check spelling of a word.

        This method takes a word in the dictionary language and returns
        `True` if it is correctly spelled, and `False` otherwise.
        """
        self._check_this()
        # Enchant asserts that the word is non-empty.
        # Check it up-front to avoid nasty warnings on stderr.
        if len(word) == 0:
            raise ValueError("can't check spelling of empty string")
        val = _e.dict_check(self._this, word.encode())
        if val == 0:
            return True
        if val > 0:
            return False
        self._raise_error()

    def suggest(self, word: str) -> List[str]:
        """Suggest possible spellings for a word.

        This method tries to guess the correct spelling for a given
        word, returning the possibilities in a list.
        """
        self._check_this()
        # Enchant asserts that the word is non-empty.
        # Check it up-front to avoid nasty warnings on stderr.
        if len(word) == 0:
            raise ValueError("can't suggest spellings for empty string")
        suggs = _e.dict_suggest(self._this, word.encode())
        return [w.decode() for w in suggs]

    def add(self, word: str) -> None:
        """Add a word to the user's personal word list."""
        self._check_this()
        _e.dict_add(self._this, word.encode())

    def remove(self, word: str) -> None:
        """Add a word to the user's personal exclude list."""
        self._check_this()
        _e.dict_remove(self._this, word.encode())

    def add_to_pwl(self, word: str) -> None:
        """Add a word to the user's personal word list."""
        warnings.warn(
            "Dict.add_to_pwl is deprecated, please use Dict.add",
            category=DeprecationWarning,
            stacklevel=2,
        )
        self._check_this()
        _e.dict_add_to_pwl(self._this, word.encode())

    def add_to_session(self, word: str) -> None:
        """Add a word to the session personal list."""
        self._check_this()
        _e.dict_add_to_session(self._this, word.encode())

    def remove_from_session(self, word: str) -> None:
        """Add a word to the session exclude list."""
        self._check_this()
        _e.dict_remove_from_session(self._this, word.encode())

    def is_added(self, word: str) -> bool:
        """Check whether a word is in the personal word list."""
        self._check_this()
        return _e.dict_is_added(self._this, word.encode())

    def is_removed(self, word: str) -> bool:
        """Check whether a word is in the personal exclude list."""
        self._check_this()
        return _e.dict_is_removed(self._this, word.encode())

    def store_replacement(self, mis: str, cor: str) -> None:
        """Store a replacement spelling for a miss-spelled word.

        This method makes a suggestion to the spellchecking engine that the
        miss-spelled word `mis` is in fact correctly spelled as `cor`.  Such
        a suggestion will typically mean that `cor` appears early in the
        list of suggested spellings offered for later instances of `mis`.
        """
        if not mis:
            raise ValueError("can't store replacement for an empty string")
        if not cor:
            raise ValueError("can't store empty string as a replacement")
        self._check_this()
        _e.dict_store_replacement(self._this, mis.encode(), cor.encode())

    store_replacement._DOC_ERRORS = ["mis", "mis"]  # type: ignore

    def __describe(self, check_this: bool = True) -> Tuple[str, str, str, str]:
        """Return a tuple describing the dictionary.

        This method returns a four-element tuple describing the underlying
        spellchecker system providing the dictionary.  It will contain the
        following strings:

            * language tag
            * name of dictionary provider
            * description of dictionary provider
            * dictionary file

        Direct use of this method is not recommended - instead, access this
        information through the attributes :py:attr:`tag` and :py:attr:`provider`.
        """
        if check_this:
            self._check_this()
        _e.dict_describe(self._this, self.__describe_callback)
        return self.__describe_result

    def __describe_callback(self, tag, name, desc, file):
        """Collector callback for dictionary description.

        This method is used as a callback into the `_enchant` function
        `enchant_dict_describe`.  It collects the given arguments in
        a tuple and stores them in the attribute :py:attr:`__describe_result`.
        """
        tag = tag.decode()
        name = name.decode()
        desc = desc.decode()
        file = file.decode()
        self.__describe_result = (tag, name, desc, file)


class DictWithPWL(Dict):
    """Dictionary with separately-managed personal word list.

    .. note::
        As of version 1.4.0, enchant manages a per-user pwl and
        exclude list.  This class is now only needed if you want
        to explicitly maintain a separate word list in addition to
        the default one.

    This class behaves as the standard class :py:class:`Dict`, but also manages a
    personal word list stored in a separate file.  The file must be
    specified at creation time by the `pwl` argument to the constructor.
    Words added to the dictionary are automatically appended to the pwl file.

    A personal exclude list can also be managed, by passing another filename
    to the constructor in the optional `pel` argument.  If this is not given,
    requests to exclude words are ignored.

    If either `pwl` or `pel` are `None`, an in-memory word list is used.
    This will prevent calls to :py:meth:`add()` and :py:meth:`remove()` from affecting the user's
    default word lists.

    The `Dict` object managing the PWL is available as the :py:attr:`pwl` attribute.
    The `Dict` object managing the PEL is available as the :py:attr:`pel` attribute.

    To create a `DictWithPWL` from the user's default language, use `None`
    as the `tag` argument.
    """

    _DOC_ERRORS = ["pel", "pel", "PEL", "pel"]

    def __init__(
        self, tag: str, pwl: str = None, pel: str = None, broker: Broker = None
    ) -> None:
        """DictWithPWL constructor.

        The argument `pwl`, if not `None,` names a file containing the
        personal word list.  If this file does not exist, it is created
        with default permissions.

        The argument `pel`, if not `None,` names a file containing the personal
        exclude list.  If this file does not exist, it is created with
        default permissions.
        """
        super().__init__(tag, broker)
        if pwl is not None:
            if not os.path.exists(pwl):
                f = open(pwl, "wt")
                f.close()
                del f
            self.pwl = self._broker.request_pwl_dict(pwl)
        else:
            self.pwl = PyPWL()
        if pel is not None:
            if not os.path.exists(pel):
                f = open(pel, "wt")
                f.close()
                del f
            self.pel = self._broker.request_pwl_dict(
                pel
            )  # type: Union[None, _e.t_dict, PyPWL]
        else:
            self.pel = PyPWL()

    def _check_this(self, msg: Optional[str] = None) -> None:
        """Extend :py:meth:`Dict._check_this()` to check PWL validity."""
        if self.pwl is None:
            self._free()
        if self.pel is None:
            self._free()
        super()._check_this(msg)
        self.pwl._check_this(msg)
        self.pel._check_this(msg)

    def _free(self) -> None:
        """Extend :py:meth:`Dict._free()` to free the PWL as well."""
        if self.pwl is not None:
            self.pwl._free()
            self.pwl = None
        if self.pel is not None:
            self.pel._free()
            self.pel = None
        super()._free()

    def check(self, word: str) -> bool:
        """Check spelling of a word.

        This method takes a word in the dictionary language and returns
        `True` if it is correctly spelled, and `False` otherwise.  It checks
        both the dictionary and the personal word list.
        """
        if self.pel.check(word):
            return False
        if self.pwl.check(word):
            return True
        if super().check(word):
            return True
        return False

    def suggest(self, word: str) -> List[str]:
        """Suggest possible spellings for a word.

        This method tries to guess the correct spelling for a given
        word, returning the possibilities in a list.
        """
        suggs = super().suggest(word)
        suggs.extend([w for w in self.pwl.suggest(word) if w not in suggs])
        for i in range(len(suggs) - 1, -1, -1):
            if self.pel.check(suggs[i]):
                del suggs[i]
        return suggs

    def add(self, word: str) -> None:
        """Add a word to the associated personal word list.

        This method adds the given word to the personal word list, and
        automatically saves the list to disk.
        """
        self._check_this()
        self.pwl.add(word)
        self.pel.remove(word)

    def remove(self, word: str) -> None:
        """Add a word to the associated exclude list."""
        self._check_this()
        self.pwl.remove(word)
        self.pel.add(word)

    def add_to_pwl(self, word: str) -> None:
        """Add a word to the associated personal word list.

        This method adds the given word to the personal word list, and
        automatically saves the list to disk.
        """
        self._check_this()
        self.pwl.add_to_pwl(word)
        self.pel.remove(word)

    def is_added(self, word: str) -> bool:
        """Check whether a word is in the personal word list."""
        self._check_this()
        return self.pwl.is_added(word)

    def is_removed(self, word: str) -> bool:
        """Check whether a word is in the personal exclude list."""
        self._check_this()
        return self.pel.is_added(word)


##  Create a module-level default broker object, and make its important
##  methods available at the module level.
_broker = Broker()
request_dict = _broker.request_dict
request_pwl_dict = _broker.request_pwl_dict
dict_exists = _broker.dict_exists
list_dicts = _broker.list_dicts
list_languages = _broker.list_languages
get_param = _broker.get_param
set_param = _broker.set_param


#  Expose the "get_version" function.
def get_enchant_version() -> str:
    """Get the version string for the underlying enchant library."""
    return _e.get_version().decode()


#  Expose the "set_prefix_dir" function.
def set_prefix_dir(path: str) -> None:
    """Set the prefix used by the Enchant library to find its plugins

    Called automatically when the Python library is imported when
    required.
    """
    return _e.set_prefix_dir(path)

    set_prefix_dir._DOC_ERRORS = ["plugins"]


def get_user_config_dir() -> str:
    """Return the path that will be used by some
    Enchant providers to look for custom dictionaries.
    """
    return _e.get_user_config_dir().decode()
